/****************************************************************************************** 
 *	Chili DirectX Framework Version 16.07.20											  *	
 *	Game.cpp																			  *
 *	Copyright 2016 PlanetChili.net <http://www.planetchili.net>							  *
 *																						  *
 *	This file is part of The Chili DirectX Framework.									  *
 *																						  *
 *	The Chili DirectX Framework is free software: you can redistribute it and/or modify	  *
 *	it under the terms of the GNU General Public License as published by				  *
 *	the Free Software Foundation, either version 3 of the License, or					  *
 *	(at your option) any later version.													  *
 *																						  *
 *	The Chili DirectX Framework is distributed in the hope that it will be useful,		  *
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of						  *
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the						  *
 *	GNU General Public License for more details.										  *
 *																						  *
 *	You should have received a copy of the GNU General Public License					  *
 *	along with The Chili DirectX Framework.  If not, see <http://www.gnu.org/licenses/>.  *
 ******************************************************************************************/
#include "MainWindow.h"
#include "Game.h"

Game::Game( MainWindow& wnd )
	:
	wnd( wnd ),
	gfx( wnd )
{
	m_grid = std::make_unique<HexGrid>(Vec2i(10.0f, 10.0f), Vec2f(64.0f, 64.0f), 32.0f);
}

void Game::Go()
{
	gfx.BeginFrame();	
	UpdateModel();
	ComposeFrame();
	gfx.EndFrame();
}

void Game::UpdateModel()
{
	static bool leftPressed = false;
	if (wnd.mouse.LeftIsPressed() && !leftPressed)
	{
		HexCell* cell = m_grid->getNodeMouse(Vec2f(wnd.mouse.GetPosX(), wnd.mouse.GetPosY()));
		if (cell)
		{
			cell->color = Colors::Red;
			cell->passable = false;
		}
	}
	leftPressed = wnd.mouse.LeftIsPressed();
	if (wnd.kbd.KeyIsPressed(VK_SPACE))
	{
		m_pathPts = m_grid->getPath(Vec2i(0, 0), Vec2i(9, 9));
	}
}

void Game::ComposeFrame()
{
	m_grid->Draw(gfx);
	if (m_pathPts.size() > 1)
	{
		for (int c = 0; c < m_pathPts.size() - 1; c++)
		{
			gfx.DrawLine(m_pathPts[c].x, m_pathPts[c].y, m_pathPts[c + 1].x, m_pathPts[c + 1].y, Colors::Blue);
		}
	}
}
