#pragma once
#include "vec2.h"
#include <vector>
#include "Graphics.h"
#include <d3d11.h>
#include <DirectXMath.h>
#define DEAD_VALUE_FLOAT -196719671967.0f
template<typename T>
static int __compareTo(T a, T b)
{
	if (a == b)
		return 0;
	if (a > b)
		return 1;

	return -1;// a < b
}
struct HexCell
{
	Vec2f center;
	float size;
	int gCost;
	int hCost;
	int fCost() { return gCost + hCost; }
	int heapIndex = 0;
	bool passable = true;
	Color color = Colors::Green;
	float lineSize = 1;
	std::vector<Vec2f> pts;
	Vec2i indicies;
	struct HexCell* parent;
	int compareTo(struct HexCell* next)
	{
		int compare = __compareTo<int>(fCost(), next->fCost());

		if (compare == -1)
		{
			compare = 1;
		}
		else if (compare == 1)
		{
			compare = -1;
		}
		if (compare == 0)
		{
			if (__compareTo<int>(hCost, next->hCost) == -1)
				compare = 1;
			else
				compare = -1;
		}
		return compare;
	}
	HexCell()
	{
		pts.resize(6);
	}
	bool pointIn(float x, float y)
	{
		const float q2x = fabsf(x - center.x);
		const float q2y = fabsf(y - center.y);
		float _horz = fabsf(pts[1].y - center.y);
		float _vert = fabsf(pts[2].y - center.y);
		if (q2x > _horz || q2y > _vert * 2)
			return false;
		return 2 * _vert * _horz - _vert * q2x - _horz * q2y >= 0;


	}
};
class HexGrid
{
private:
	class Heap
	{
		int currItemCount = 0;
		int maxItems;
		std::vector<HexCell*> items;
		void sortDown(HexCell* item)
		{
			while (true)
			{
				int indexLeft = item->heapIndex * 2 + 1;
				int indexRight = item->heapIndex * 2 + 2;
				int swapIndex = 0;
				if (indexLeft < currItemCount)
				{
					swapIndex = indexLeft;
					if (indexRight < currItemCount)
					{
						if (items[indexLeft]->compareTo(items[indexRight]) < 0)
						{
							swapIndex = indexRight;
						}
					}
					if (item->compareTo(items[swapIndex]) < 0)
					{
						swap(item, items[swapIndex]);
					}
					else
					{
						return;
					}
				}
				else
				{
					return;
				}

			}
		}
		void sortUp(HexCell* item)
		{
			int parentIndex = (item->heapIndex - 1) / 2;
			while (true)
			{
				HexCell* parentItem = items[parentIndex];
				if (item->compareTo(parentItem) > 0)
				{
					swap(item, parentItem);
				}
				else
				{
					break;
				}
				parentIndex = (item->heapIndex - 1) / 2;
			}
		}
		void swap(HexCell* itemA, HexCell* itemB)
		{
			items[itemA->heapIndex] = itemB;
			items[itemB->heapIndex] = itemA;
			int temp = itemA->heapIndex;
			itemA->heapIndex = itemB->heapIndex;
			itemB->heapIndex = temp;
		}
	public:
		Heap() = default;
		Heap(int maxItems)
		{
			this->maxItems = maxItems;
			items.resize(maxItems);
		}
		void reset()
		{
			items.clear();
			items.resize(maxItems);
			currItemCount = 0;
		}
		HexCell* removeFirst()
		{
			HexCell* n = items[0];
			currItemCount--;
			items[0] = items[currItemCount];
			items[0]->heapIndex = 0;
			sortDown(items[0]);
			return n;
		}
		void updateItem(HexCell* item)
		{
			sortUp(item);
		}
		int count()
		{
			return currItemCount;
		}
		bool contains(HexCell* item)
		{
			return items[item->heapIndex] == item;
		}
		void add(HexCell* item)
		{
			item->heapIndex = currItemCount;
			items[currItemCount] = item;
			sortUp(item);
			currItemCount++;
		}
	};
private:
	int columns;
	int rows;
	Heap* pHeap = nullptr;
	Vec2f nodeSize;
	std::vector<HexCell*> cells;
	std::vector<HexCell*> closed_list;
	Vec2f getPoint(Vec2f center, float size, int i)
	{
		float angle_deg = -180.0f + (60.0f * (float)i + 30.0f);
		float angle_rad = DirectX::XM_PI / 180.0f * angle_deg;
		Vec2f Point(center.x + size * cos(angle_rad),
			center.y + size * sin(angle_rad));
		return Point;
	}
	int getDistance(HexCell* nodeA, HexCell* nodeB)
	{
		int distX = abs(nodeA->center.x - nodeB->center.x);
		int distY = abs(nodeA->center.y - nodeB->center.y);

		if (distX > distY)
			return 14 * distY + 10 * (distX - distY);

		return 14 * distX + 10 * (distY - distX);
	}
	void addToClosedList(HexCell* node)
	{
		if (!isInClosedList(node))
			closed_list.push_back(node);
	}
	bool isInClosedList(HexCell* node)
	{
		auto& it = std::find(closed_list.begin(), closed_list.end(), node);
		if (it != closed_list.end())
		{
			return true;
		}
		return false;
	}
public:
	HexGrid() = default;
	HexGrid(Vec2i rc, Vec2f startPos, float _size)
	{

		nodeSize = Vec2f(_size, _size);
		Vec2f center = startPos;
		float size = _size;
		rows = rc.y;
		columns = rc.x;
		pHeap = new Heap(rows*columns);
		for (int i = 0; i < rows; i++)
		{
			if (i % 2 == 1)
			{
				center.x = size + (sqrtf(3.0f) / 2.0f * (2 * size)) * 0.5f;
			}
			else
			{
				center.x = size;
			}

			for (int j = 0; j < columns; j++)
			{
				// calculate the points in the hex
				HexCell* cell = new HexCell();
				for (int c = 0; c < 6; c++)
				{

					cell->center = center;
					cell->size = size;
					cell->pts[c] = getPoint(center, size, c);
					cell->indicies = Vec2i(j, i);

				}
				cells.push_back(cell);

				center.x += sqrtf(3.0f) / 2.0f * (size * 2.0f);
			}
			center.y += (size * 2) * 0.75f;
		}



	}
	~HexGrid()
	{
		for (auto& it : cells)
			delete it;

		delete pHeap;
	}
	void  Draw(Graphics& gfx)
	{


		for (auto& it : cells)
		{

			for (int i = 0; i < 5; i++)
			{

				gfx.DrawLine(it->pts[i].x,it->pts[i].y ,it->pts[i + 1].x,it->pts[i + 1].y , it->color);
			}
			gfx.DrawLine( it->pts[5].x,it->pts[5].y ,
			 it->pts[0].x,it->pts[0].y , it->color);
			
		}


	};
	std::vector<Vec2f> retraceSteps(HexCell* start, HexCell* end)
	{
		std::vector<Vec2f> positions;

		HexCell* curr = end;
		while (curr != start)
		{
			positions.push_back(curr->center);
			curr = curr->parent;
		}

		positions.push_back(start->center);
		std::reverse(positions.begin(), positions.end());
		return positions;
	}
	std::vector<HexCell*> getAdjacentCells(HexCell* node)
	{
		Vec2i pos = node->indicies;
		std::vector<HexCell*> result;
		bool odd;
		int x, y;
		(pos.y % 2 == 1) ? odd = true : odd = false;

		if (odd)// odd row
		{
			// top
			x = pos.x + 0;
			if ((y = pos.y - 1) >= 0) // top
			{
				result.push_back(cells[y * columns + x]);
				if ((x = pos.x + 1) < columns) // right
					result.push_back(cells[y * columns + x]);
			}

			// middle row
			y = pos.y;
			if ((x = pos.x - 1) >= 0) // left
				result.push_back(cells[y * columns + x]);
			if ((x = pos.x + 1) < columns) // right
				result.push_back(cells[y * columns + x]);

			//bottom row
			if ((y = pos.y + 1) < rows)
			{
				if ((x = pos.x + 1) < columns)
					result.push_back(cells[y * columns + x]);
				x = pos.x;
				result.push_back(cells[y * columns + x]);
			}

		}
		else // even row
		{
			// top	
			if ((y = pos.y - 1) >= 0) // in top bounds
			{

				if ((x = pos.x - 1) >= 0) // left bounds
					result.push_back(cells[y * columns + x]);
				x = pos.x;
				result.push_back(cells[y * columns + x]);
			}
			// middle
			y = pos.y;
			if ((x = pos.x - 1) >= 0) // left bounds
				result.push_back(cells[y * columns + x]);
			;
			if ((x = pos.x + 1) < columns) // right bounds
				result.push_back(cells[y * columns + x]);

			// bottom
			if ((y = pos.y + 1) < rows) // in bottom bounds
			{
				x = pos.x;
				result.push_back(cells[y * columns + x]);

				if ((x = pos.x - 1) >= 0) // left bounds
					result.push_back(cells[y * columns + x]);
			}

		}
		return result;
	}
	std::vector<Vec2f> getPath(Vec2i start, Vec2i end)
	{
		pHeap->reset();

		HexCell* startNode = getNode(start);
		HexCell* endNode = getNode(end);
		// if invalid position(s)
		std::vector<Vec2f> dead_find;
		dead_find.emplace_back(Vec2f(DEAD_VALUE_FLOAT, DEAD_VALUE_FLOAT));
		if (startNode == endNode)
			return dead_find;

		if (!startNode || !endNode)
			return dead_find;

		if ((startNode && !startNode->passable) || (endNode && !endNode->passable))
			return dead_find;


		closed_list.clear();
		pHeap->add(startNode);
		while (pHeap->count() > 0u)
		{
			HexCell* curr_node = pHeap->removeFirst();
			addToClosedList(curr_node);
			if (curr_node == endNode)
			{
				return retraceSteps(startNode, endNode);
			}
			std::vector<HexCell*> neighbour = getAdjacentCells(curr_node);
			for (auto& it : neighbour)
			{
				// eliminate these in "getNeightbours" ??? 
				if (!it->passable || isInClosedList(it))
					continue;

				// find cheapest cost
				const int newMoveCost = curr_node->gCost + getDistance(curr_node, it);
				bool inOpenList = pHeap->contains(it);
				if (newMoveCost < it->gCost || !inOpenList)
				{
					// we are here so add and set cheapest node
					it->gCost = newMoveCost;
					it->hCost = getDistance(it, endNode);
					it->parent = curr_node;
					if (!inOpenList)
						pHeap->add(it);
					else
						pHeap->updateItem(it);
				}

			}
		}
		return dead_find;
	}
	HexCell* getNodeMouse(Vec2f pos)
	{

		for (int i = 0; i< cells.size(); i++)
		{
			if (cells[i]->pointIn(pos.x, pos.y))
			{
				return cells[i];
			}
		}
		return nullptr;
	}
	HexCell* getNode(Vec2i pos)
	{
		int row = pos.y;
		int col = pos.x;
		int index = row * columns + col;

		if (index < 0 || index >= cells.size())
			return nullptr;

		return cells[index];
	}
};